@@ -12,20 +12,19 @@ use MVC::Neaf;
 my $err_tpl = <<"TT";
 <html><head><title>Error [% error %] - demo</title></head>
 <body><h1>[% message %]</h1>
-<h2>Propagated in [% caller.1 %]</h2>
 <a href="/">Back to safety...</a>
 TT
 
-MVC::Neaf->error_template( 404 => {
+MVC::Neaf->set_error_handler( 404 => {
     -template => \$err_tpl,
     message => "You are searching in the wrong place",
 } );
-MVC::Neaf->error_template( 405 => {
+MVC::Neaf->set_error_handler( 405 => {
     -template => \$err_tpl,
     message => "Not the right method to access this page",
 } );
 
-MVC::Neaf->route( cgi => basename(__FILE__) => sub {} => method => []
+MVC::Neaf->route( cgi => basename(__FILE__) => sub {} => method => ['HEAD']
     => description => "Demonstrate custom error template" );
 
 MVC::Neaf->run;
