@@ -12,6 +12,6 @@ use MVC::Neaf;
 # Add some flexibility to run alongside other examples
 my $script = basename(__FILE__);
 
-MVC::Neaf->static( "/examples" => $Bin, buffer => 2 );
+MVC::Neaf->static( "/examples" => $Bin, buffer => 1024, dir_index=>1 );
 
 MVC::Neaf->run;
