@@ -9,16 +9,12 @@ my $Bin;
 BEGIN { $Bin = dirname( __FILE__ ) || "." };
 use lib $Bin."/../lib";
 use MVC::Neaf;
-use MVC::Neaf::X::Session::File;
+use MVC::Neaf::X::Session::Cookie;
 
 my $script = basename(__FILE__);
-my $storage = "$Bin/nocommit-$script-storage";
-
-mkdir $storage; # ignore result
--d $storage or die "Failed to find directory '$storage' ($!)";
 
 MVC::Neaf->set_session_handler(
-    engine => MVC::Neaf::X::Session::File->new( dir => $storage ),
+    engine => MVC::Neaf::X::Session::Cookie->new( key => 'foobared' ),
     view_as => 'session'
 );
 
