@@ -4,7 +4,7 @@ use 5.006;
 use strict;
 use warnings;
 
-our $VERSION = 0.16;
+our $VERSION = 0.17;
 
 =head1 NAME
 
@@ -119,6 +119,9 @@ It will be executed AFTER the headers and pre-generated content
 are served to the client, and may use C<$req-E<gt>write( $data );>
 and C<$req-E<gt>close;> to output more data.
 
+=item * -headers - Pass a hash or array of values for header generation.
+This is an alternative to L<MVC::Neaf::Request>'s C<push_header> method.
+
 =item * -jsonp - Used by JS view module as a callback name to produce a
 L<jsonp|https://en.wikipedia.org/wiki/JSONP> response.
 Callback MUST be a set of identifiers separated by dots.
@@ -160,11 +163,12 @@ use Carp;
 use Scalar::Util qw(blessed looks_like_number);
 use Encode;
 use URI::Escape;
+use Module::Load;
 use parent qw(Exporter);
 
-our @EXPORT_OK = qw(neaf_err neaf get post);
+our @EXPORT_OK = qw( neaf_err neaf get post head put );
 our %EXPORT_TAGS = (
-    sugar => [qw[neaf get post]],
+    sugar => [qw[ neaf get post head put ]],
 );
 
 use MVC::Neaf::Util qw(http_date canonize_path path_prefixes run_all run_all_nodie);
@@ -199,15 +203,11 @@ methods do not intersect.
 HEAD method is automatically handled if GET is present, however,
 one MAY define a separate HEAD handler explicitly.
 
-=item * path_info_regex => qr/.../ - allow URI subpaths matching
-the specified expression to be handled by this handler.
+=item * path_info_regex => qr/.../ - allow URI subpaths
+to be handled by this handler.
 
-If specified, 404 error will be returned unless PATH_INFO matches the regex
-(without the leading slash).
-
-Starting from v.0.16, this parameter is going to be REQUIRED for C<path_info()>
-method to be used in the handler.
-Until then, a DEPRECATED warning will be generated for a naked path_info call.
+A 404 error will be generated unless C<path_info_regex> is present
+and PATH_INFO matches the regex (without the leading slash).
 
 B<EXPERIMENTAL>. Name and semantics MAY change in the future.
 
@@ -253,6 +253,10 @@ sub route {
     my (%args) = @_;
     $self = $Inst unless ref $self;
 
+    # check defaults to be a hash before accessing them
+    $self->_croak( "default must be unblessed hash" )
+        if $args{default} and ref $args{default} ne 'HASH';
+
     # minus-prefixed keys are typically defaults
     $_ =~ /^-/ and $args{default}{$_} = delete $args{$_}
         for keys %args;
@@ -277,9 +281,8 @@ sub route {
     $profile{caller}   = [caller(0)]; # file,line
     if ( !defined $args{path_info_regex} ) {
         # TODO replace def path_info_regex with '' in v.0.16 aka '404 if unspecified'
-        $args{path_info_regex}      = '.*';
-    } else {
-        $profile{has_path_info_regex} = 1;
+        $args{path_info_regex}      = '';
+        $profile{no_path_info_regex} = 1;
     };
     $profile{path_info_regex}  = qr#^/*($args{path_info_regex})$#;
 
@@ -288,23 +291,15 @@ sub route {
     $profile{description} = $args{description};
 
     if (my $view = $args{view}) {
-        if (!ref $view) {
-            $view = $self->load_view( $view );
-        } elsif (ref $view eq 'CODE') {
-            $view = MVC::Neaf::View->new( on_render => $view );
-        };
+        # preload view so we can fail early
+        $view = $self->get_view( $view );
 
-        $self->_croak( "view must be a coderef or a MVC::Neaf::View object" )
-            unless blessed $view and $view->isa("MVC::Neaf::View");
-
-        $profile{view} = $view;
+        $args{default}{-view} = ref $args{view} ? $view : $args{view};
     };
 
-    if (my $def = $args{default}) {
-        $self->_croak( "default must be unblessed hash" )
-            unless ref $def eq 'HASH';
-        $profile{todo_default} = $args{default};
-    };
+    # todo_default because some path-based defs will be mixed in later
+    $profile{todo_default} = $args{default}
+        if $args{default};
 
     if ( $args{cache_ttl} ) {
         $self->_croak("cache_ttl must be a number")
@@ -320,9 +315,31 @@ sub route {
     $self->{route}{ $path }{$_} = { %profile, my_method => $_ }
         for @{ $args{method} };
 
+    # This is for get+post sugar
+    $self->{last_added} = \%profile;
+
     return $self;
 }; # end sub route
 
+# This is for get+post sugar
+# TODO merge with alias, GET => implicit HEAD
+sub _dup_route {
+    my ($self, $method, $profile) = @_;
+
+    $profile ||= $self->{last_added};
+    my $path = $profile->{path};
+
+    $self->_croak( "Attempting to set duplicate handler for [$method] "
+        .($path || '/'))
+            if $self->{route}{ $path }{$method};
+
+    $self->{route}{ $path }{$method}
+        and $self->_croak("Attempting to set duplicate ");
+
+    delete $self->{route_re};
+    $self->{route}{ $path }{$method} = { %$profile, my_method => $method };
+};
+
 =head2 alias( $newpath => $oldpath )
 
 Create a new name for already registered route.
@@ -358,11 +375,6 @@ sub alias {
 
 Serve static content located under C<$file_path>.
 
-File type detection is based on extention.
-This MAY change in the future.
-Known file types are listed in C<%MVC::Neaf::X::Files::ExtType> hash.
-Patches welcome.
-
 %options may include:
 
 =over
@@ -378,12 +390,34 @@ B<EXPERIMENTAL>. Cache API is not yet established.
 =item * allow_dots => 1|0 - if true, serve files/directories
 starting with a dot (.git etc), otherwise give a 404.
 
-B<EXPERIMAENTAL>
+B<EXPERIMENTAL>
+
+=item * dir_index => 1|0 - if true, generate index for a directory;
+otherwise a 404 is returned, and deliberately so, for security reasons.
+
+B<EXPERIMENTAL>
+
+=item * dir_template - specify template for directory listing
+(with images etc). A sane default is provided.
+
+B<EXPERIMENTAL>
+
+=item * view - specify view object for rendering dir template.
+By default a localized TT instance is used.
+
+B<EXPERIMENTAL> Name MAY be changed (dir_view etc).
 
 =item * description - comment. The default is "Static content at $dir"
 
 =back
 
+The content is really handled by L<MVC::Neaf::X::Files>.
+
+File type detection is based on extention.
+This MAY change in the future.
+Known file types are listed in C<%MVC::Neaf::X::Files::ExtType> hash.
+Patches welcome.
+
 Generally it is probably a bad idea to serve files in production
 using a web application framework.
 Use a real web server instead.
@@ -394,23 +428,14 @@ This is the intended usage.
 
 =cut
 
-my %static_options;
-$static_options{$_}++ for qw( description buffer cache_ttl allow_dots );
 sub static {
     my ($self, $path, $dir, %options) = @_;
     $self = $Inst unless ref $self;
 
-    my @extra = grep { !$static_options{$_} } keys %options;
-    $self->_croak( "Unknown options @extra" )
-        if @extra;
-
     require MVC::Neaf::X::Files;
-    my $xfiles = MVC::Neaf::X::Files->new( %options, root => $dir );
-    return $self->route($path => $xfiles->make_handler
-        , method => ['GET', 'HEAD']
-        , path_info_regex => '.*'
-        , cache_ttl => $options{cache_ttl}
-        , description => $options{description} || "Static content at $dir" );
+    my $xfiles = MVC::Neaf::X::Files->new(
+        %options, root => $dir, base_url => $path );
+    return $self->route( $xfiles->make_route );
 };
 
 =head2 pre_route( sub { ... } )
@@ -435,13 +460,11 @@ sub pre_route {
     return $self;
 };
 
-=head2 load_view( $view_name, [ $object || $module_name, %options ] )
+=head2 load_view( $name, $object || coderef || ($module_name, %options) )
 
-Fetch existing, if any, or set up a new view object named C<$view_name>.
-
-This will be called whenever the app returns { -view => "foo" }.
-
-If no view was found, the following rules apply:
+Load a view object and cache it under name $name, if $name is true.
+The loaded view is returned.
+All subsequent calls to get_view( $name ) would return that object, too.
 
 =over
 
@@ -457,21 +480,6 @@ Those are prefixed with C<MVC::Neaf::View::>.
 
 If C<set_forced_view> was called, return its argument instead.
 
-So the intended usage is as follows:
-
-    # in the app initialisation section
-    # this can be omitted when using TT, JS, or Dumper as view.
-    MVC::Neaf->load_view( foo => 'TT', INCLUDE_PATH => "/foo/bar" );
-        # Short alias for MVC::Neaf::View::TT
-
-    # in the app itself
-    return {
-        ...
-        -view => "foo", # triggers a second call with 1 parameter "foo"
-    };
-
-B<NOTE> Some convoluted logic here, needs rework.
-
 =cut
 
 my %known_view = (
@@ -480,39 +488,37 @@ my %known_view = (
     Dumper => 'MVC::Neaf::View::Dumper',
 );
 sub load_view {
-    my ($self, $view, $module, @param) = @_;
+    my ($self, $name, $obj, @param) = @_;
     $self = $Inst unless ref $self;
 
+    $self->_croak("At least two arguments required")
+        unless defined $name and defined $obj;
+
+    # We've been overridden!
     return $self->{force_view}
         if exists $self->{force_view};
-    $view = $self->{-view}
-        unless defined $view;
 
-    # Already an object - just return it after the checks
-    return $view if ref $view;
-    # Agressive caching FTW!
-    return $self->{seen_view}{$view}
-        if exists $self->{seen_view}{$view};
-
-    # If we got to this point, we don't have $view cached just yet.
-    # Instantiate and save!
-
-    if (!ref $module) {
-        # If 1-arg call is used, try using view name as class name
-        $module ||= $view;
+    # Instantiate if needed
+    if (!ref $obj) {
         # in case an alias is used, apply alias
-        $module = $known_view{ $module } || $module;
+        $obj = $known_view{ $obj } || $obj;
 
         # Try loading...
-        eval "require $module"; ## no critic
-        $self->_croak( "Failed to load view $view: $@" )
-            if $@;
-        $module = $module->new( @param );
+        if (!$obj->can("new")) {
+            eval { load $obj; 1 }
+                or $self->_croak( "Failed to load view $name=>$obj: $@" );
+        };
+        $obj = $obj->new( @param );
     };
 
-    $self->{seen_view}{$view} = $module;
+    $self->_croak( "view must be a coderef or a MVC::Neaf::View object" )
+        unless blessed $obj and $obj->can("render")
+            or ref $obj eq 'CODE';
 
-    return $module;
+    $self->{seen_view}{$name} = $obj
+        if $name;
+
+    return $obj;
 };
 
 =head2 set_default ( key => value, ... )
@@ -619,18 +625,28 @@ sub set_session_handler {
     my ($self, %opt) = @_;
     $self = $Inst unless ref $self;
 
-    my $sess = $opt{engine};
-    my $cook = $opt{cookie} || 'session';
+    my $sess = delete $opt{engine};
+    my $cook = $opt{cookie} || 'neaf.session';
 
     $self->_croak("engine parameter is required")
         unless $sess;
+
+    if (!ref $sess) {
+        $opt{session_ttl} = delete $opt{ttl} || $opt{session_ttl};
+
+        my $obj = eval { load $sess; $sess->new( %opt ); }
+            or $self->_croak("Failed to load session '$sess': $@");
+
+        $sess = $obj;
+    };
+
     my @missing = grep { !$sess->can($_) }
         qw(get_session_id session_id_regex session_ttl
             create_session load_session save_session delete_session );
     $self->_croak("engine object does not have methods: @missing")
         if @missing;
 
-    my $regex = $sess->session_id_regex || qr(^$);
+    my $regex = $sess->session_id_regex;
     my $ttl   = $opt{ttl} || $sess->session_ttl || 0;
 
     $self->{session_handler} = [ $sess, $cook, $regex, $ttl ];
@@ -696,6 +712,7 @@ B<DEPRECATED>. Same as above, but issues a warning.
 
 =cut
 
+# TODO kill in v.0.20
 sub error_template {
     my $self = shift;
 
@@ -881,25 +898,44 @@ It is not stable yet, so be careful when upgrading Neaf.
 Create a route with C<GET/HEAD> methods enabled.
 The %options are the same as those of C<route()> method.
 
-=cut
+=head2 head '/path' => CODE, %options;
 
-sub get(@) { ## no critic # DSL
-    my ($path, $handler, @args) = @_;
-
-    return MVC::Neaf->route( $path, $handler, method => [ 'GET', 'HEAD' ], @args );
-};
+Create a route with C<HEAD> method enabled.
+The %options are the same as those of C<route()> method.
 
 =head2 post '/path' => CODE, %options;
 
 Create a route with C<POST> method enabled.
 The %options are the same as those of C<route()> method.
 
+=head2 put '/path' => CODE, %options;
+
+Create a route with C<PUT> method enabled.
+The %options are the same as those of C<route()> method.
+
+=head2 get + post '/path' => CODE, %options;
+
+B<EXPERIMENTAL>. Set multiple methods in one go.
+
 =cut
 
-sub post(@) { ## no critic # DSL
-    my ($path, $handler, @args) = @_;
+foreach (qw(get head post put)) {
+    my $method = uc $_;
 
-    return MVC::Neaf->route( $path, $handler, method => [ 'POST' ], @args );
+    my $code = sub(@) { ## no critic
+        # get + post sugar
+        if (@_ == 1 and UNIVERSAL::isa( $_[0], __PACKAGE__ )) {
+            return $_[0]->_dup_route( $method );
+        };
+
+        # normal operation
+        my ($path, $handler, @args) = @_;
+
+        return $Inst->route( $path, $handler, @args, method => $method );
+    };
+
+    no strict 'refs'; ## no critic
+    *{$_} = $code;
 };
 
 =head2 neaf->...
@@ -963,6 +999,10 @@ sub neaf(@) { ## no critic # DSL
         $action = 'hook';
     };
 
+    if ($action eq 'session') {
+        unshift @args, 'engine';
+    };
+
     my $method = $method_shortcut{$action};
     croak "neaf: don't know how to handle '$action'"
         unless $method and MVC::Neaf->can($method);
@@ -970,7 +1010,7 @@ sub neaf(@) { ## no critic # DSL
     return MVC::Neaf->$method( @args );
 };
 
-=head1 EXPERIMENTAL HOOKS
+=head1 HOOKS
 
 Hooks are subroutines executed during various phases of request processing.
 Each hook is characterized by phase, code to be executed, path, and method.
@@ -978,17 +1018,10 @@ Multiple hooks MAY be added for the same phase/path/method combination.
 ALL hooks matching a given route will be executed, either short to long or
 long to short (aka "event bubbling"), depending on the phase.
 
-B<CAUTION> Avoid hooks if possible.
-Hook handling code is the most convoluted, complex part of the Neaf CORE.
-
-Use normal subroutine calls,
-even if copied-and-pasted across multiple controllers.
-Use guard objects if you want automatic resource deallocation.
-Subclass L<MVC::Neaf::View> and/or its children if you want to add an action
-around content rendering.
-
-Only use hooks when other options are exhausted,
-or there's a clear, straightforward use case.
+B<CAUTION> Don't overuse hooks.
+This may lead to a convoluted, hard to follow application.
+Use hooks for repeated auxiliary tasks such as checking permissions or writing
+down statistics, NOT for primary application logic.
 
 =head2 add_hook ( phase => CODEREF, %options )
 
@@ -1024,14 +1057,42 @@ Multiple locations may be supplied via C<[ /foo, /bar ...]>
 =item * method => 'METHOD' || [ list ]
 List of request HTTP methods to which given hook applies.
 
-=item * prepend => 0|1 - if true, move hook execution to earlier stage.
-Note that this does NOT override path precedence.
+=item * prepend => 0|1 - all other parameters being equal,
+hooks will be executed in order of adding.
+This option allows to override this and run given hook first.
+Note that this does NOT override path bubbling order.
 
 =back
 
 =head2 HOOK PHASES
 
 This list of phases MAY change in the future.
+Current request processing diagram looks as follows:
+
+   [*] request created
+    . <- pre_route [no path] [can die]
+    |
+    * route - select handler
+    |
+    . <- pre_logic [can die]
+   [*] execute main handler
+    * apply path-based defaults - reply() is populated now
+    |
+    . <- pre_content
+    ? checking whether content already generated
+    |\
+    | . <- pre_render [can die - template error produced]
+    | [*] render - -content is present now
+    |/
+    * generate default headers (content type & length, cookies, etc)
+    . <- pre_reply [path traversal long to short]
+    |
+   [*] headers sent out, no way back!
+    * output the rest of reply (if -continue specified)
+    * execute postponed actions (if any)
+    |
+    . <- pre_cleanup [path traversal long to short] [no effect on headers]
+   [*] request destroyed
 
 =head3 pre_route
 
@@ -1041,7 +1102,7 @@ resolved and handler found.
 Dying in this phase stops both further hook processing and controller execution.
 Instead, the corresponding error handler is executed right away.
 
-Setting C<path> and C<exclude> is not available on this stage.
+Options C<path> and C<exclude> are not available on this stage.
 
 May be useful for mangling path.
 Use C<$request-E<gt>set_full_path($new_path)> if you need to.
@@ -1202,7 +1263,7 @@ sub set_forced_view {
     delete $self->{force_view};
     return $self unless $view;
 
-    $self->{force_view} = $self->load_view( $view );
+    $self->{force_view} = $self->get_view( $view );
 
     return $self;
 };
@@ -1326,7 +1387,7 @@ sub handle_request {
         };
         $path_info =~ $route->{path_info_regex}
             or die "404\n";
-        $req->set_full_path( $path, $path_info, !$route->{has_path_info_regex} );
+        $req->set_full_path( $path, $path_info, $route->{no_path_info_regex} );
         $self->_post_setup( $route )
             unless exists $route->{lock};
 
@@ -1339,7 +1400,7 @@ sub handle_request {
 
     if ($data) {
         # post-process data - fill in request(RD) & global(GD) defaults.
-        # TODO fill in per-location defaults, too - but do we need them?
+        # TODO kill request defaults in v.0.20
         my $RD = $req->get_default;
         my $GD = $route->{default};
         exists $data->{$_} or $data->{$_} = $RD->{$_} for keys %$RD;
@@ -1365,15 +1426,17 @@ sub handle_request {
     # produce headers later.
     my $content = \$data->{-content};
     if( !defined $$content) {
-        my $view = $self->load_view( $data->{-view} || $route->{view} );
+        my $view = $self->get_view( $data->{-view} || $route->{view} );
         eval {
             run_all( $route->{hooks}{pre_render}, $req )
                 if exists $route->{hooks}{pre_render};
-            ($$content, my $type) = $view->render( $data );
+            ($$content, my $type) = blessed $view
+                ? $view->render( $data ) : $view->( $data );
             $data->{-type} ||= $type;
         };
         if (!defined $$content) {
             # TODO $req->clear; - but don't kill cleanup hooks
+            # FIXME bug here - resetting data does NOT affect the inside of req
             $self->_log_error( view => $@ );
             $data = {
                 -status => 500,
@@ -1403,8 +1466,19 @@ sub handle_request {
             unless $data->{-type} =~ /; charset=/;
     };
 
-    # Mangle headers - these modifications remain stored in req
+    # Mangle headers - NOTE these modifications remain stored in req
     my $head = $req->header_out;
+    if (my $append = $data->{-headers}) {
+        if (ref $append eq 'ARRAY') {
+            for (my $i = 0; $i < @$append; $i+=2) {
+                $head->push_header($append->[$i], $append->[$i+1]);
+            };
+        }
+        else {
+            carp("Neaf: -headers must be ARRAY, not ".(ref $append));
+            # Would love to die, but it's impossible here
+        };
+    };
     $head->init_header( content_type => $data->{-type} || $self->{-type} );
     $head->init_header( location => $data->{-location} )
         if $data->{-location};
@@ -1562,6 +1636,27 @@ sub _log_error {
     warn $msg;
 };
 
+=head2 get_view( "name" )
+
+Fetch view object by name.
+Uses C<load_view> w/o additional params if needed.
+This is for internal usage.
+
+=cut
+
+sub get_view {
+    my ($self, $view) = @_;
+    $self = $Inst unless ref $self;
+
+    if (ref $view) {
+        return $self->load_view( '' => $view );
+    }
+    else {
+        $view = $self->{-view} unless defined $view;
+        return $self->{seen_view}{$view} || $self->load_view( $view, $view );
+    };
+};
+
 =head2 run_test( \%PSGI_ENV, %options )
 
 =head2 run_test( "/path?param=value", %options )
@@ -1569,6 +1664,8 @@ sub _log_error {
 Run a PSGI request and return a list of
 C<($status, HTTP::Headers, $whole_content )>.
 
+Returns just the content in scalar context.
+
 Just as the name suggests, useful for testing only (it reduces boilerplate).
 
 Continuation responses are supported.
@@ -1577,8 +1674,12 @@ Continuation responses are supported.
 
 =over
 
+=item * method - set method (default is GET)
+
 =item * override = \%hash - force certain data in ENV
 
+=item * cookie = \%hash - force HTTP_COOKIE header
+
 =back
 
 =cut
@@ -1599,8 +1700,16 @@ sub run_test {
         }
     };
     # TODO more civilized stuff like cookies, headers...
+    $env->{REQUEST_METHOD} = $opt{method} if $opt{method};
     $env->{$_} = $opt{override}{$_} for keys %{ $opt{override} };
 
+    if (my $cook = $opt{cookie}) {
+        # TODO hash processing
+        $env->{HTTP_COOKIE} = $env->{HTTP_COOKIE}
+            ? "$env->{HTTP_COOKIE}, $cook"
+            : $cook;
+    };
+
     my $ret = $self->run->( $env );
     if (ref $ret eq 'CODE') {
         # PSGI functional interface used.
