@@ -3,7 +3,7 @@ package MVC::Neaf::Request;
 use strict;
 use warnings;
 
-our $VERSION = 0.16;
+our $VERSION = 0.17;
 
 =head1 NAME
 
@@ -214,6 +214,67 @@ sub script_name {
     return $self->{script_name};
 };
 
+=head2 get_url_base()
+
+Get scheme, server, and port of the application.
+
+B<EXPERIMENTAL> Name and meaning subject to change.
+
+=head2 get_url_rel( %override )
+
+Produce a relative link to the page being served,
+possibly overriding some parameters.
+
+Parameter order is NOT preserved. If parameter is empty or undef,
+it is skipped.
+
+B<CAUTION> Multi-values are ignored, this MAY change in the future.
+
+B<CAUTION> For a POST request, normal parameters are used instead of URL
+parameters (see C<url_param>). This MAY change in the future.
+
+B<EXPERIMENTAL> Name and meaning subject to change.
+
+=head2 get_url_full( %override )
+
+Same as above, but prefixed with schema, server name, and port.
+
+B<EXPERIMENTAL> Name and meaning subject to change.
+
+=cut
+
+sub get_url_rel {
+    my ($self, %override) = @_;
+
+    my %h = (%{ $self->_all_params }, %override );
+
+    return $self->path . '?' . join '&'
+        , map { uri_escape_utf8($_) . "=" .uri_escape_utf8($h{$_}) }
+        grep { defined $h{$_} && length $h{$_} }
+        sort keys %h;
+};
+
+my %port_scheme = (
+    http => 80,
+    https => 443,
+);
+
+sub get_url_base {
+    my $self = shift;
+
+    # skip well-known ports
+    my $port = ($self->port == ($port_scheme{ $self->scheme } || 0))
+        ? ''
+        : ':'.$self->port;
+
+    return join "", $self->scheme, "://", $self->hostname, $port;
+};
+
+sub get_url_full {
+    my $self = shift;
+    return $self->get_url_base . $self->get_url_rel(@_);
+};
+
 =head2 path_info()
 
 Returns the part of URI path beyond what matched the application's path.
@@ -222,28 +283,22 @@ Contrary to the
 L<CGI specification|https://tools.ietf.org/html/rfc3875#section-4.1.5>,
 the leading slash is REMOVED.
 
-The validation regexp for this value SHOULD be specified during application
+The validation regexp for this value MUST be specified during application
 setup as C<path_info_regex>. See C<route> in L<MVC::Neaf>.
 
-B<NOTE> Starting v.0.16 of this module, path_info() will die unless
-validation regexp was provided.
-
 B<NOTE> Experimental. This part of API is undergoing changes.
 
 =cut
 
 sub path_info {
-    my ($self, $regexp) = @_;
+    my ($self) = @_;
 
     if ($self->{no_path_info_regex}) {
         # TODO all instances of no_path_info_regex must be killed in v.0.16,
         # and undefined path_info die here
-        carp "DEPRECATED path_info() called, but path_info_regex validation was not set in route()";
-    } elsif (defined $regexp) {
-        carp "DEPRECATED path_info() called with regex, use path_info_regex parameter in route() instead";
+        $self->_croak( "path_info() called, but path_info_regex validation was not set in route()" );
     };
 
-
     return $self->{path_info};
 };
 
@@ -464,43 +519,30 @@ sub form {
 
 =head2 get_form_as_hash ( name => qr/.../, name2 => qr/..../, ... )
 
-Return a group of form parameters as a hashref.
-Only values that pass corresponding validation are added.
-
-B<DEPRECATED>. Use L<MVC::Neaf::X::Form> instead.
+B<DEPRECATED> and dies. Use L<MVC::Neaf::X::Form> instead.
 
 =cut
 
 sub get_form_as_hash {
     my ($self, %spec) = @_;
 
-    carp "DEPRECATED. get_form_as_hash() will be removed soon.";
-
-    my %form;
-    foreach (keys %spec) {
-        my $value = $self->param( $_, $spec{$_}, undef );
-        $form{$_} = $value if defined $value;
-    };
-
-    return \%form;
+    $self->_croak( "DEPRECATED. use MVC::Neaf::X::Form and form() method" );
 };
 
 =head2 get_form_as_list ( qr/.../, qw(name1 name2 ...)  )
 
 =head2 get_form_as_list ( [ qr/.../, "default" ], qw(name1 name2 ...)  )
 
-Return a group of form parameters as a list, in that order.
+Return a group of uniform parameters as a list, in that order.
 Values that fail validation are returned as undef, unless default given.
 
-B<DEPRECATED>. Use L<MVC::Neaf::X::Form> instead.
+B<EXPERIMENTAL>. The name MAY be changed in the future.
 
 =cut
 
 sub get_form_as_list {
     my ($self, $spec, @list) = @_;
 
-    carp "DEPRECATED. get_form_as_list() will be removed soon.";
-
     $self->_croak( "Meaningless call in scalar context" )
         unless wantarray;
 
@@ -544,17 +586,20 @@ sub body {
 =head2 set_default( key => $value, ... )
 
 Set default values for your return hash.
-May be useful inside MVC::Neaf->pre_route.
+May be useful inside C<pre_route> and C<pre_logic> hooks.
 
 Returns self.
 
-B<EXPERIMANTAL>. API and naming subject to change.
+B<DEPRECATED>. Use path-based defaults and/or $request->stash instead.
+This will be removed in v.0.20+.
 
 =cut
 
 sub set_default {
     my ($self, %args) = @_;
 
+    # TODO remove in v.0.20
+    carp "Neaf:set_default() DEPRECATED. Use path-based defaults or stash()";
     foreach (keys %args) {
         defined $args{$_}
             ? $self->{defaults}{$_} = $args{$_}
@@ -580,6 +625,29 @@ sub get_default {
 
 =head2 upload( "name" )
 
+Returns an L<MVC::Neaf::Upload> object corresponding to an uploaded file,
+if such uploaded file is present.
+
+An upload object has at least C<handle> and C<content> methods to work with
+data:
+
+    my $upload = $req->upload("user_file");
+    if ($upload) {
+        my $untrusted_filename = $upload->filename;
+        my $fd = $upload->handle;
+        while (<$fd>) {
+            ...
+        };
+    }
+
+or just
+
+    if ($upload) {
+        while ($upload->content =~ /(...)/g) {
+            do_something($1);
+        };
+    };
+
 =cut
 
 sub upload {
@@ -1120,12 +1188,18 @@ sub _set_reply {
 
 =head2 stash()
 
+=head2 stash( "name" )
+
+=head2 stash( %save_data )
+
 A hashref that is guaranteed to persist throughout the request lifetime.
 
 This may be useful to maintain shared data accross hooks and callbacks.
 
 Use C<session> if you intend to share data between requests.
+
 Use C<reply> if you intend to render the data for the user.
+
 Use C<stash> as a last resort for temporary, private data.
 
 Stash is not killed by C<clear()> function so that cleanup isn't
@@ -1139,7 +1213,14 @@ too cumbersome.
 sub stash {
     my $self = shift;
     my $st = $self->{stash} ||= {};
-    return $st;
+    return $st unless @_;
+
+    return $st->{ $_[0] } unless @_>1;
+
+    $self->_croak("Odd number of elements in hash assignment") if @_ % 2;
+    my %new = @_;
+    $st->{$_} = $new{$_} for keys %new;
+    return $self;
 };
 
 =head2 postpone( CODEREF->(req) )
